'This object handles the chess board
'Copyright 2008 Raymond Allen


CON
    'number of custom 16x16 characters              
  nuchars = 13*9+8    '!!!!! you must have the correct # here for alignment later

  'Piece Colors
  White=0  'Because White goes first
  Black=1
  nColors=2
  '  // Codes representing pieces
  PAWN = 0
  KNIGHT = 2
  BISHOP = 4
  ROOK = 6
  QUEEN = 8
  KING = 10
  ExtraKing = 12
  EnPassant = 14
  AllPieces = 16
  NoPiece = 18
  'nRegularPieces = 12 
  WHITE_PAWN = PAWN + WHITE
  WHITE_KNIGHT = KNIGHT + WHITE
  WHITE_BISHOP = BISHOP + WHITE
  WHITE_ROOK = ROOK + WHITE
  WHITE_QUEEN = QUEEN + WHITE
  WHITE_KING = KING + WHITE
  White_ExtraKing = ExtraKing + White
  White_EnPassant = EnPassant + White  
  White_AllPieces = AllPieces + White
  BLACK_PAWN = PAWN + BLACK
  BLACK_KNIGHT = KNIGHT + BLACK
  BLACK_BISHOP = BISHOP + BLACK
  BLACK_ROOK = ROOK + BLACK
  BLACK_QUEEN = QUEEN + BLACK
  BLACK_KING = KING + BLACK
  Black_ExtraKing = ExtraKing + Black
  Black_EnPassant = EnPassant + Black 
  Black_AllPieces = AllPieces + Black
  'bitboard constants
  nBitBoards=18

  'Move constants 
  'Move Structure (byte withing long)
  MOVE_SourceSquare=0
  MOVE_DestinationSquare=1
  MOVE_MovedPiece=2
  MOVE_Type=3
  MOVE_CapturedPiece=4
  MOVE_Bytes=5  '# of bytes in move structure
  '// The different types of moves recognized by the game (one byte)
  MOVE_NORMAL = 0
  MOVE_CAPTURE = 1
  MOVE_EN_PASSANT = 2
  MOVE_CASTLING_KINGSIDE = 4
  MOVE_CASTLING_QUEENSIDE = 8
  MOVE_RESIGN = 15
  MOVE_STALEMATE = 14
  MOVE_PROMOTION_KNIGHT = 16
  MOVE_PROMOTION_BISHOP = 32
  MOVE_PROMOTION_ROOK = 64
  MOVE_PROMOTION_QUEEN = 128
  '// A pair of masks used to split the promotion and the non-promotion part of a move type ID
  PROMOTION_MASK = %1111_0000
  NO_PROMOTION_MASK = %0000_1111
  
OBJ

VAR
  'for custom characters
  word user_charbase
  'bitboards:  'need 2 longs for each (64-bit) bitboard 
  long BitBoard[nBitBoards<<1]  'The entire state of the game (aside from whose turn it is) is contained within this array
  long SquareBitBoard[2]  'used for converting square to bitboard  
  'game vars
  Byte Player  'Whose turn it is
  long MaterialValue[nColors]
  byte NumPawns[nColors]
  'move vars
  byte CurrentMove[MOVE_Bytes]

  


PUB Start
  '64 byte align the user characters
  user_charbase := @uchar & $FFC0 'destination
  'user_charbase_offset := user_charbase-@uchar  
  longmove(user_charbase,@uchar,16*nuchars)

PRI EmptyBoard 'clear all bitboards
  longfill(@BitBoard,0,nBitBoards<<1)

Pub Init:bOK|i
  'Setup the board to play
   'clear running totals
  repeat i from white to black
    MaterialValue[i]:=0
    NumPawns[i]:=0

   'Put the pieces on the board
  EmptyBoard
  AddPiece( 0, BLACK_ROOK )
  AddPiece( 1, BLACK_KNIGHT )
  AddPiece( 2, BLACK_BISHOP )
  AddPiece( 3, BLACK_QUEEN )
  AddPiece( 4, BLACK_KING )
  AddPiece( 5, BLACK_BISHOP )
  AddPiece( 6, BLACK_KNIGHT )
  AddPiece( 7, BLACK_ROOK )
  repeat  i from 8 to 15
    AddPiece( i, BLACK_PAWN )
  repeat i from 3 to 5
    AddPiece(i, Black_ExtraKing)

  repeat i from 48 to 55
    AddPiece( i, WHITE_PAWN )
  AddPiece( 56, WHITE_ROOK )
  AddPiece( 57, WHITE_KNIGHT )
  AddPiece( 58, WHITE_BISHOP )
  AddPiece( 59, WHITE_QUEEN )
  AddPiece( 60, WHITE_KING )
  AddPiece( 61, WHITE_BISHOP )
  AddPiece( 62, WHITE_KNIGHT )
  AddPiece( 63, WHITE_ROOK )
  repeat i from 59 to 61
    AddPiece(i, White_ExtraKing)

   'White to play the first move
  Player:=White
  return true  

PUB MakeMove(StartSquare, EndSquare):pErrorString|pMove
  pMove:=@CurrentMove
  byte[pMove][MOVE_SourceSquare]:=StartSquare
  byte[pMove][MOVE_DestinationSquare]:=EndSquare
  '// Time to try to figure out what the move means!
  'if ( Player==WHITE ) 'combining black&white logic
    ' Is there a piece (of the moving player) on SourceSquare?
    ' If not, abort
    byte[pMove][MOVE_MovedPiece]:= GetPiece(StartSquare,White+Player)
    if ( byte[pMove][MOVE_MovedPiece] == NoPiece )
      return String("Sorry, You don't have a piece on that square")
   
    'Three cases: there is a piece on the destination square (a capture),
    'the destination square allows an en passant capture, or it is a
    'simple non-capture move.  If the destination contains a piece of the
    'moving side, abort
    if ( GetPiece(EndSquare,White+Player) <> NoPiece )
      return String("Sorry, can't capture your own piece!" )


    'Capture?
    byte[pMove][MOVE_CapturedPiece]:= GetPiece(EndSquare,Black-Player)
    if (byte[pMove][MOVE_CapturedPiece] <> NoPiece)
      byte[pMove][MOVE_Type]:= MOVE_Normal
    elseif ( BitOnSquare(EnPassant+Black-Player, byte[pMove][MOVE_DestinationSquare]  ) and ( byte[pMove][MOVE_MovedPiece] == WHITE_PAWN+Player ) )
      byte[pMove][MOVE_CapturedPiece]:= BLACK_PAWN-Player
      byte[pMove][MOVE_Type]:= MOVE_EN_PASSANT
 
    ' If the move isn't a capture, it may be a castling attempt
    elseif ( ( byte[pMove][MOVE_MovedPiece] == WHITE_KING+Player ) AND ( ( byte[pMove][MOVE_SourceSquare] - byte[pMove][MOVE_DestinationSquare] ) == 2 ) )
      byte[pMove][MOVE_Type]:= MOVE_CASTLING_KINGSIDE
    elseif ( ( byte[pMove][MOVE_MovedPiece] == WHITE_KING+Player ) AND ( ( byte[pMove][MOVE_SourceSquare] - byte[pMove][MOVE_DestinationSquare] ) == -2 ) )
      byte[pMove][MOVE_Type]:= MOVE_CASTLING_QUEENSIDE
    else
      byte[pMove][MOVE_Type]:= MOVE_NORMAL

  'Pawn Promotion?
  if (((byte[pMove][MOVE_MovedPiece] == WHITE_PAWN+Player) and (byte[pMove][MOVE_DestinationSquare] < 8)) or ((Byte[pMove][MOVE_MovedPiece] == BLACK_PAWN-Player) and (byte[pMove][MOVE_DestinationSquare] > 55)))
    byte[pMove][MOVE_Type] += MOVE_PROMOTION_QUEEN
    pErrorString:=String("Promotion.")  'need to get exact promotion choice from user...


PRI BitOnSquare(Piece, Square):bOnSquare
  'check given bitboard for a set bit on given square
  if square<32
    return BitBoard[Piece<<1] | (1<<Square)
  else
    return BitBoard[Piece<<1+1] | (1<<(Square-32))  

PUB ApplyMove(StartSquare, EndSquare)  

PRI AddPiece(Square, Piece): bAdded|SquareBits0, SquareBits1, Color
  'init
  SquareBits0:=0
  SquareBits1:=0
  Color:=Piece//2
   
  'Add a piece to a bitboard
  if square<32
    SquareBits0:= 1<<Square
    BitBoard[Piece<<1] |= SquareBits0
  else
    SquareBits1:= 1<<(Square-32) 
    BitBoard[Piece<<1+1] |= SquareBits1
   
  ' And note the new piece position in the bitboard containing all
  ' pieces of its color.  
  BitBoard[(White_AllPieces + Color)<<1] |= SquareBits0
  BitBoard[(White_AllPieces + Color)<<1+1] |= SquareBits1
   
   
  ' And adjust material balance accordingly
  MaterialValue[Color] += PieceValue[Piece]
  if ( Piece == WHITE_PAWN )
    NumPawns[White]++
  elseif ( Piece == BLACK_PAWN )
    NumPawns[Black]++

PRI bPieceOnSquare(Piece, Square)|SquareBits
  'Look for piece on it's bitboard bitboard
  if square<32
    SquareBits:= 1<<Square
    if BitBoard[Piece<<1] & SquareBits
      return true
  else
    SquareBits:= 1<<(Square-32) 
    if BitBoard[Piece<<1+1] & SquareBits
      return true
  return false  

PUB GetSquareBitmap(square):BmpAddress|p
'Get the bitmap that should be painted on given square
  p:=@Blank  
  case GetPiece(square,White)
    WHITE_PAWN:
      p:=@WhitePawn 
    WHITE_KNIGHT:
      p:=@WhiteKnight
    WHITE_BISHOP:
      p:=@WhiteBishop
    WHITE_ROOK:
      p:=@WhiteRook  
    WHITE_QUEEN:
      p:=@WhiteQueen
    WHITE_KING:
      p:=@WhiteKing

  case GetPiece(square,Black)
    BLACK_PAWN:
      p:=@BlackPawn 
    BLACK_KNIGHT:
      p:=@BlackKnight
    BLACK_BISHOP:
      p:=@BlackBishop
    BLACK_ROOK:
      p:=@BlackRook  
    BLACK_QUEEN:
      p:=@BlackQueen
    BLACK_KING:
      p:=@BlackKing

  return BmpAddress:=p+user_charbase-@uchar

PUB GetBorderBitmap(n):BmpAddress|p
  case n
    0: p:=@border_top
    1: p:=@border_right
    2: p:=@border_bottom
    3: p:=@border_left
    4: p:=@border_top_left
    5: p:=@border_top_right
    6: p:=@border_bottom_right
    7: p:=@border_bottom_left

  return BmpAddress:=p+user_charbase-@uchar  

PRI GetPiece(square,color)
'Scan bitboards for a White piece on this square
  if bPieceOnSquare(WHITE_KING+color, square) 
    return WHITE_KING+color
  if bPieceOnSquare(WHITE_QUEEN+color, square)
    return WHITE_QUEEN+color
  if bPieceOnSquare(WHITE_ROOK+color, square)
    return WHITE_ROOK+color
  if bPieceOnSquare(WHITE_KNIGHT+color, square)
    return WHITE_KNIGHT+color
  if bPieceOnSquare(WHITE_BISHOP+color, square)
    return WHITE_BISHOP+color
  if bPieceOnSquare(WHITE_PAWN+color, square)
    return WHITE_PAWN+color
  return NoPiece

DAT 'initialized data
  PieceValue byte 10,10,30,30,35,35,50,50,90,90,200,200   ' Numerical evaluation of piece material values
  
'DAT
padding LONG  7[16] 'alignment padding for the following user defined characters     

uchar long
'pieces are 2-bit bitmaps.  There are 13 types, each with nine 16x16 tiles. Size is 13*9*(256*2/8)  =7.5kB
WhitePawn long
        file "pawn_white3.dat"  '3x3=9
BlackPawn long
        file "pawn_black3.dat"  '3x3=9
WhiteRook long
        file "rook_white3.dat"  '3x3=9       
BlackRook long
        file "rook_black3.dat"  '3x3=9
WhiteKnight long
        file "Knight_white3.dat"  '3x3=9       
BlackKnight long
        file "Knight_black3.dat"  '3x3=9
WhiteBishop long
        file "Bishop_white3.dat"  '3x3=9       
BlackBishop long
        file "Bishop_black3.dat"  '3x3=9
WhiteQueen long
        file "Queen_white3.dat"  '3x3=9       
BlackQueen long
        file "Queen_black3.dat"  '3x3=9
WhiteKing long
        file "King_white3.dat"  '3x3=9       
BlackKing long
        file "King_black3.dat"  '3x3=9
Blank long 0[144]  '9
Border_top long
        file "border_top.dat"  '1x1=1
Border_right long
        file "border_right.dat" '1x1=1
border_bottom long
        file "border_bottom.dat" '1x1
border_left long
        file "border_left.dat"  '1x1

Border_top_left long
        file "border_top_left.dat"  '1x1=1
Border_top_right long
        file "border_top_right.dat" '1x1=1
border_bottom_right long
        file "border_bottom_right.dat" '1x1
border_bottom_left long
        file "border_bottom_left.dat"  '1x1








        